import json

import pytest

from copy import deepcopy

from jsonschema.exceptions import SchemaError
from santoku.utils.configuration import (
    ConfigurationError,
    ConfigurationManager,
    IllegalAccessPattern,
    SchemaViolation,
    SettingError,
)


@pytest.fixture(scope="function")
def valid_settings_A():
    return {
        "integer_setting": 1,
        "boolean_setting": True,
        "string_setting": "initial string value",
        "float_setting": 1.0,
        "dict_setting": {
            "nested_string_setting": "initial dict string value",
            "nested_boolean_setting": False,
        },
    }


@pytest.fixture(scope="function")
def valid_configuration_A(valid_settings_A):
    return {"name": "valid_configuration_A", "settings": valid_settings_A}


@pytest.fixture(scope="function")
def valid_settings_B():
    return {
        "integer_setting": 2,
        "boolean_setting": False,
        "string_setting": "test string value",
        "float_setting": 2.0,
        "dict_setting": {
            "nested_string_setting": "test dict string value",
            "nested_boolean_setting": True,
        },
    }


@pytest.fixture(scope="function")
def valid_configuration_B(valid_settings_B):
    return {"name": "valid_configuration_B", "settings": valid_settings_B}


@pytest.fixture(scope="function")
def schema():
    return {
        "type": "object",
        "properties": {
            "integer_setting": {"type": "integer"},
            "boolean_setting": {"type": "boolean"},
            "string_setting": {"type": "string"},
            "float_setting": {"type": "number"},
            "dict_setting": {
                "type": "object",
                "properties": {
                    "nested_string_setting": {"type": "string"},
                    "nested_boolean_setting": {"type": "boolean"},
                },
                "additionalProperties": False,
                "required": ["nested_string_setting", "nested_boolean_setting"],
            },
        },
        "additionalProperties": False,
        "required": [
            "integer_setting",
            "boolean_setting",
            "string_setting",
            "float_setting",
            "dict_setting",
        ],
    }


@pytest.fixture(scope="function")
def schema_with_spec():
    def get_schema_with_spec(draft: int):
        return {"$schema": f"http://json-schema.org/draft-0{draft}/schema#"}

    return get_schema_with_spec


@pytest.fixture(scope="function")
def schema_with_invalid_spec():
    return {"$schema": f"This is not a valid JSON Schema specification identifier"}


@pytest.fixture(scope="function")
def schema_violating_configuration():
    return {
        "name": "invalid_configuration",
        "settings": {},
    }


@pytest.fixture(scope="function")
def create_json_file(tmpdir):
    # creates a JSON file inside the tmpdir (provided by pytest)
    # tmpdir guarantees that any file inside it will be deleted automatically after the test
    # thus, no need to add a finalizer
    def _create_json_file(rel_path, content):
        path = tmpdir.join(rel_path)
        with open(path, "w") as f:
            json.dump(content, f)
        return path

    return _create_json_file


@pytest.fixture(scope="function")
def configurations_file(valid_configuration_A, valid_configuration_B, create_json_file):
    file_name = "configurations.json"
    configurations = [valid_configuration_A, valid_configuration_B]
    abs_path = create_json_file(rel_path=file_name, content=configurations)
    return abs_path


@pytest.fixture(scope="function")
def schema_file(tmpdir, schema, create_json_file):
    file_name = "schema.json"
    abs_path = create_json_file(rel_path=file_name, content=schema)
    return abs_path


@pytest.fixture(scope="function")
def configuration_manager(valid_configuration_A, valid_configuration_B, schema):
    configurations = [valid_configuration_A, valid_configuration_B]
    return ConfigurationManager(
        configurations=configurations,
        initial_configuration=valid_configuration_A["name"],
        schema=schema,
    )


class TestConfigurationManager:
    def test_init(
        self, valid_configuration_A, valid_configuration_B, schema, schema_violating_configuration
    ):
        # Initialize two configurations. Success expected.
        try:
            configuration_manager = ConfigurationManager(
                configurations=[valid_configuration_A, valid_configuration_B],
                schema=schema,
                initial_configuration=valid_configuration_A["name"],
            )
        except:
            assert False

        # Active configuration must be initial configuration. Success expected.
        expected_configuration_name = valid_configuration_A["name"]
        obtained_configuration_name = configuration_manager.active_configuration
        assert obtained_configuration_name == expected_configuration_name

        # Retrieve the stored settings after init. Success expected.
        expected_settings = valid_configuration_A["settings"]
        obtained_settings = configuration_manager.configurations[valid_configuration_A["name"]]
        assert obtained_settings == expected_settings

        expected_settings = valid_configuration_B["settings"]
        obtained_settings = configuration_manager.configurations[valid_configuration_B["name"]]
        assert obtained_settings == expected_settings

        # Initialize with a config that does not follow the schema. Failure expected.
        with pytest.raises(SchemaViolation):
            ConfigurationManager(configurations=[schema_violating_configuration], schema=schema)

        # Pass an initial configuration to an empty configuration manager. Failure expected.
        with pytest.raises(ConfigurationError):
            ConfigurationManager(initial_configuration=valid_configuration_A["name"])

    def test_from_json(self, configurations_file, schema_file):
        # Initialize a ConfigurationManager from a correct JSON file. Success expected.
        try:
            configuration_manager = ConfigurationManager.from_json(
                configurations_file_path=configurations_file, schema_file_path=schema_file,
            )
        except:
            assert False

    def test_check_schema(self, schema_with_spec, schema_with_invalid_spec):
        # We only check that the appropriate validators are used. The actual work of validating is
        # performed by the jsonschema package, which has its own set of tests.

        # Check valid schemas using supported specifications. Success expected.
        ConfigurationManager.check_schema(schema=schema_with_spec(draft=7))
        ConfigurationManager.check_schema(schema=schema_with_spec(draft=6))
        ConfigurationManager.check_schema(schema=schema_with_spec(draft=4))

        # Check valid schema without specifying the spec (uses the default). Success expected.
        ConfigurationManager.check_schema(schema={})

        # Check valid schemas with invalid/unsupported specs. Failure expected.
        with pytest.raises(SchemaError):
            ConfigurationManager.check_schema(schema=schema_with_invalid_spec)
        with pytest.raises(SchemaError):
            ConfigurationManager.check_schema(schema=schema_with_spec(237))

    def test_validate_schema(self, valid_settings_A, configuration_manager, schema):
        # Validate settings with less arguments than schema requires. Failure expected.
        settings_to_validate = deepcopy(valid_settings_A)
        removed_value = settings_to_validate.pop("integer_setting")
        with pytest.raises(SchemaViolation):
            configuration_manager.validate_schema(settings=settings_to_validate)

        # Validate settings with more arguments than the schema. Failure expected.
        settings_to_validate = deepcopy(valid_settings_A)
        settings_to_validate["new_setting"] = "new_value"
        with pytest.raises(SchemaViolation):
            configuration_manager.validate_schema(settings=settings_to_validate)

        # Validate settings with a differently named setting than the schema. Failure expected.
        settings_to_validate = deepcopy(valid_settings_A)
        settings_to_validate["new_integer_setting"] = settings_to_validate.pop("integer_setting")
        with pytest.raises(SchemaViolation):
            configuration_manager.validate_schema(settings=settings_to_validate)

        # Validate settings with a differently typed value than the schema. Failure expected.
        settings_to_validate = deepcopy(valid_settings_A)
        settings_to_validate["integer_setting"] = "1"
        with pytest.raises(SchemaViolation):
            configuration_manager.validate_schema(settings=settings_to_validate)

        # Validate nested setting with a different types than in the schema. Failure expected.
        settings_to_validate = deepcopy(valid_settings_A)
        settings_to_validate["dict_setting"]["nested_string_setting"] = 1
        with pytest.raises(SchemaViolation):
            configuration_manager.validate_schema(settings=settings_to_validate)

        # Validate settings that follow the schema. Success expected.
        settings_to_validate = deepcopy(valid_settings_A)
        try:
            configuration_manager.validate_schema(settings=settings_to_validate)
        except:
            assert False

    def test_define_configuration(self, valid_settings_A, schema, schema_violating_configuration):
        configuration_manager = ConfigurationManager(schema=schema)

        # Define a configuration in an empty configuration manager. Success expected.
        try:
            configuration_manager.define_configuration(
                name="configuration_name", settings=valid_settings_A
            )
        except:
            assert False

        # Configuration settings must be preserved upon insertion. Success expected.
        expected_settings = deepcopy(valid_settings_A)
        obtained_setting = configuration_manager.configurations["configuration_name"]
        assert obtained_setting == expected_settings

        # Override an existing configuration. Success expected.
        new_settings = deepcopy(valid_settings_A)
        new_settings["integer_setting"] = new_settings["integer_setting"] + 1
        configuration_manager.define_configuration(
            name="configuration_name", settings=new_settings, override=True,
        )
        obtained_settings = configuration_manager.configurations["configuration_name"]
        assert obtained_settings == new_settings

        # Define a configuration with settings that do not follow the schema. Failure expected.
        with pytest.raises(SchemaViolation):
            configuration_manager.define_configuration(
                name=schema_violating_configuration["name"],
                settings=schema_violating_configuration["settings"],
            )

        # Define an existent configuration without the override flag. Failure expected.
        with pytest.raises(ConfigurationError):
            configuration_manager.define_configuration(
                name="configuration_name", settings=valid_settings_A, override=False,
            )

    def test_list_configurations(self, configuration_manager):
        expected = list(configuration_manager.configurations)
        assert configuration_manager.list_configurations() == expected

    def test_get_configuration(self, valid_configuration_A, configuration_manager):
        # Retrieve an existing configuration. Success expected.
        expected_settings = valid_configuration_A["settings"]
        obtained_settings = configuration_manager.get_configuration(
            name=valid_configuration_A["name"]
        )
        assert obtained_settings == expected_settings

        # Retrieve an undefined configuration. Failure expected.
        with pytest.raises(ConfigurationError):
            configuration_manager.get_configuration(name="undefined_configuration")

    def test_apply_configuration(self, valid_configuration_A, configuration_manager):
        # Applied configuration must be the active one after applying. Success expected.
        configuration_manager.apply_configuration(name=valid_configuration_A["name"])
        expected_settings = valid_configuration_A["settings"]
        obtained_settings = configuration_manager.configurations[
            configuration_manager.active_configuration
        ]
        assert obtained_settings == expected_settings

        # Apply an unavailable configuration. Failure expected.
        with pytest.raises(ConfigurationError):
            configuration_manager.apply_configuration(name="unavailable_configuration")

    def test_get_active_configuration(self, valid_configuration_A, configuration_manager):
        # Check the active configuration is the expected one. Success expected.
        expected_configuration = valid_configuration_A["settings"]
        obtained_configuration = configuration_manager.get_active_configuration()
        assert obtained_configuration == expected_configuration

    def test_get_setting(self, valid_configuration_A, configuration_manager):
        # Retrieve a setting value of the active configuration. Success expected.
        expected_setting = valid_configuration_A["settings"]["integer_setting"]
        obtained_setting = configuration_manager.get_setting("integer_setting")
        assert expected_setting == obtained_setting

        # Retrieve an undefind setting. Failure expected.
        with pytest.raises(SettingError):
            configuration_manager.get_setting("undefined_setting")

        # Retrieve nested setting. Success expected.
        expected_setting = valid_configuration_A["settings"]["dict_setting"][
            "nested_boolean_setting"
        ]
        obtained_setting = configuration_manager.get_setting(
            "dict_setting", "nested_boolean_setting"
        )
        assert expected_setting == obtained_setting

        # Retrieve a nested collection of settings (forbidden access pattern). Failure expected.
        with pytest.raises(IllegalAccessPattern):
            configuration_manager.get_setting("dict_setting")

        # Retrieve an undefind nested setting. Failure expected.
        with pytest.raises(SettingError):
            configuration_manager.get_setting("dict_setting", "nested_undefined_setting")



import json
import boto3

from base64 import b64decode
from typing import Dict, Any

from botocore.exceptions import ClientError


class SecretsManagerError(Exception):
    def __init__(self, message):
        super().__init__(message)


class SecretsManagerHandler:
    """
    Class to manage operations of AWS Secrets Manager Service (SQS).

    This class is intended to be used in other projects when the use of secrets is required.
    Secrets Manager allows the storage and retreiving of secrets in a safe way. The
    connection to the Secrets Manager service is done using the service class Client of the boto3
    library.

    """

    def __init__(self, **kwargs):
        """
        Initializes the services classes.

        Notes
        -----
            More information on the available regions: [1]

        References
        ----------
        [1] :
        https://aws.amazon.com/about-aws/global-infrastructure/regions_az/

        """
        self.client = boto3.client(service_name="secretsmanager", **kwargs)

    def get_secret_value(self, secret_name: str) -> Dict[str, Any]:
        """
        Retreive the content of the secret with name `secret_name`.

        Parameters
        ----------
        secret_name : str
            Id of secret to retreive.

        Raises
        ------
        SecretsManagerError
            If the secret cannot be decrypted correctly, if there was an error on the server side,
            if there were invalid parameters, if a parameter is not valid for the current state of
            the resource, if the secret was not found.

        Returns
        -------
        Dict[str, Any]
            The content of the secret with its correspondent key and value.

        """

        try:
            secret_value_response = self.client.get_secret_value(SecretId=secret_name)
        except ClientError as e:
            if e.response["Error"]["Code"] == "DecryptionFailureException":
                error_message = (
                    "Secrets Manager can't decrypt the protected secret using the provided KMS key."
                )
                raise SecretsManagerError(error_message)
            elif e.response["Error"]["Code"] == "InternalServiceErrorException":
                error_message = "An error occured on the server side."
                raise SecretsManagerError(error_message)
            elif e.response["Error"]["Code"] == "InvalidParameterException":
                error_message = "You provided an invalid value for a parameter."
                raise SecretsManagerError(error_message)
            elif e.response["Error"]["Code"] == "InvalidRequestException":
                error_message = "You provided a parameter value that is not valid for the current state of the resource."
                raise SecretsManagerError(error_message)
            elif e.response["Error"]["Code"] == "ResourceNotFoundException":
                error_message = "Secrets Manager can't find the resource you asked for."
                raise SecretsManagerError(error_message)
            else:
                # unknown error
                raise e
        else:
            # Depending on whether the secret is binary or string one of these fields will be populated.
            if "SecretString" in secret_value_response:
                secret_str = secret_value_response["SecretString"]
            else:
                secret_str = b64decode(secret_value_response["SecretBinary"])
            secret = json.loads(secret_str)
        return secret



import json
import re
from typing import Dict, List, Optional
from urllib import parse

import pandas as pd
import requests
from requests import HTTPError
from santoku.aws import SecretsManagerHandler


class SalesforceObjectError(Exception):
    def __init__(self, message):
        super().__init__(message)


class SalesforceObjectFieldError(Exception):
    def __init__(self, message):
        super().__init__(message)


class RequestMethodError(Exception):
    def __init__(self, message):
        super().__init__(message)


class LightningRestApiHandler:
    """
    Class to manage operations on Salesforce content.

    This class contains methods that interact with Salesforce objects and makes easy some
    usual operations. The connection is done by calling the Salesforce REST API. This class is
    pretended to be used on AWS Glue jobs (Python Shell) directly or through a higher level API.

    Notes
    -----
    The functionalities of the private methods are the following:
        Pass authentication credentials to establish connection with Salesforce.
        Collect the object names available in Salesforce, which is used to verify the correctness
        of the parameters.
        Collect the object fields available of a specific Salesforce object and store this
        information as a cache, which is used to verify the correctness of the parameters.
        Extract the object name from a given path.
        Verify if the introduced parameters are valid fields of a Salesforce object.

    More information on the use of Salesforce REST API: [1]

    References
    ----------
    [1] :
    https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/quickstart.htm

    """

    def __init__(
        self,
        auth_url: str,
        username: str,
        password: str,
        client_id: str,
        client_secret: str,
        api_version: str = "54.0",
        grant_type: str = "password",
    ) -> None:
        """
        Initialize the private variables of the class.

        Parameters
        ----------
        auth_url : str
            Url used to authenticate with Salesforce.
        username : str
            Username used to authenticate with Salesforce.
        password : str
            Password used to authenticate with Salesforce.
        client_id : str
            Consumer key used to authenticate with Salesforce.
        client_secret : str
            Consumer secret used to authenticate with Salesforce.
        api_version : str, optional
            Version of the Salesforce API used (the default is 54.0).
        grant_type : str, optional
            Type of credentials used to authenticate with Salesforce(the default is 'password').

        """
        self._url_to_format = "{}/services/data/v{}/{}"

        self._auth_url = auth_url
        self._api_version = api_version
        self._grant_type = grant_type
        self._username = username
        self._password = password
        self._client_id = client_id
        self._client_secret = client_secret

        self._instance_scheme_and_authority = ""
        self._access_token = ""

        self._salesforce_object_names_cache: List[str] = []
        self._salesforce_object_fields_cache: Dict[str, List[str]] = {}
        self._salesforce_object_required_fields_cache: Dict[str, List[str]] = {}

        self.request_headers: Dict[str, str] = {
            "Authorization": "OAuth",
            "Content-type": "application/json",
            "Accept-Encoding": "gzip",
        }

        # Indicates if there is need to validate whether the requesting Salesforce object is valid.
        self._validate_salesforce_object = True
        self._is_authenticated = False

    @classmethod
    def from_aws_secrets_manager(
        cls,
        secret_name: str,
        secret_keys: Dict[str, str] = {
            "auth_url_key": "AUTH_URL",
            "username_key": "USR",
            "password_key": "PSW",
            "client_id_key": "CLIENT_USR",
            "client_secret_key": "CLIENT_PSW",
        },
        api_version: str = "54.0",
        grant_type: str = "password",
    ) -> "LightningRestApiHandler":
        """
        Retrieve the Salesforce credentials from AWS Secrets Manager and initialize the class.
        Requires that AWS credentials with the appropriate permissions are located somewhere on the
        AWS credential chain in the local machine.

        Parameters
        ----------
        secret_name : str
            Name or ARN for the secret containing the JSON needed for the Salesforce authentication.
        secret_keys : Dict[str, str], optional
            Sepecification of the secret keys used in AWS Secrets Manager to store the credentials.
            (By default "AUTH_URL", "USR", "PSW", "CLIENT_USR", "CLIENT_PSW" will be the keys that
            stores the Salesforce credentials.)
        api_version : str, optional
            Version of the Salesforce API used (the default is 54.0).
        grant_type : str, optional
            Type of credentials used to authenticate with Salesforce(the default is 'password').

        Raises
        ------
        ValueError
            If the `secret_keys` argument does not contain the required attributes.

        See Also
        --------
        __init__ : this method calls the constructor.

        Notes
        -----
        The `secret_keys` parameter must be a JSON containing fixed attributes:
        ```
        {
            "auth_url_key": <auth url>,
            "username_key": <username>,
            "password_key": <password>,
            "client_id_key": <client id>,
            "client_secret_key": <clientsecret>,
        }
        ```

        """
        for key in [
            "auth_url_key",
            "username_key",
            "password_key",
            "client_id_key",
            "client_secret_key",
        ]:
            if key not in secret_keys:
                raise ValueError("The `secret_keys` argument does not contain the required keys.")

        secrets_manager = SecretsManagerHandler()
        credential_info = secrets_manager.get_secret_value(secret_name=secret_name)

        return cls(
            auth_url=credential_info[secret_keys["auth_url_key"]],
            username=credential_info[secret_keys["username_key"]],
            password=credential_info[secret_keys["password_key"]],
            client_id=credential_info[secret_keys["client_id_key"]],
            client_secret=credential_info[secret_keys["client_secret_key"]],
            api_version=api_version,
            grant_type=grant_type,
        )

    def _authenticate(self) -> None:
        try:
            response = requests.post(
                self._auth_url,
                data={
                    "grant_type": self._grant_type,
                    "username": self._username,
                    "password": self._password,
                    "client_id": self._client_id,
                    "client_secret": self._client_secret,
                },
                headers={"Accept": "application/json"},
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as err:
            raise
        else:
            self._is_authenticated = True

            response_as_dict = response.json()

            self._instance_scheme_and_authority = response_as_dict["instance_url"]
            self._access_token = response_as_dict["access_token"]
            # Update header with OAuth access token.
            self.request_headers["Authorization"] = f"OAuth {self._access_token}"

    def get_salesforce_object_names(self) -> List[str]:
        """
        Return the sobjects in the organization.

        Return
        ------
        List[str]
            List of names of the Salesforce object in the organization.

        Raises
        ------
        HTTPError
            If the connection with Salesforce fails, e.g. the record does not exist.

        """
        if not self._salesforce_object_names_cache:
            self._validate_salesforce_object = False

            # GET request to /sobjects returns a list with the valid objects.
            response = self.do_request(method="GET", path="sobjects")

            self._salesforce_object_names_cache = [
                sobject["name"] for sobject in json.loads(response)["sobjects"]
            ]

        return self._salesforce_object_names_cache

    def get_salesforce_object_fields(self, salesforce_object_name: str) -> List[str]:
        """
        Return the arguments that an sobject has.

        Parameters
        ----------
        salesforce_object_name : str, optional
            A Salesforce object.

        Return
        ------
        List[str]
            List of all the fields that a Salesforce object has.

        Raises
        ------
        HTTPError
            If the connection with Salesforce fails, e.g. the record does not exist.
        """
        if salesforce_object_name not in self._salesforce_object_fields_cache:
            self._validate_salesforce_object = False
            response = self.do_request(
                method="GET", path=f"sobjects/{salesforce_object_name}/describe"
            )

            self._salesforce_object_fields_cache[salesforce_object_name] = [
                fields["name"] for fields in json.loads(response)["fields"]
            ]

            # Update also the required fields to save a call to the API.
            self._salesforce_object_required_fields_cache[salesforce_object_name] = [
                fields["name"]
                for fields in json.loads(response)["fields"]
                if not fields["nillable"]
                and not fields["defaultedOnCreate"]
                and fields["createable"]
            ]

        return self._salesforce_object_fields_cache[salesforce_object_name]

    def get_salesforce_object_required_fields(self, salesforce_object_name: str) -> List[str]:
        """
        Return the mandatory arguments that an sobject needs to be created.

        Parameters
        ----------
        salesforce_object_name : str, optional
            A Salesforce object.

        Return
        ------
        List[str]
            List of the required fields that a Salesforce object needs to be created.

        Notes
        -----
        Salesforce has not made a definition of what should be `required`. However, we have found
        that some records of objects need specific fields to be given when they are created. We have
        observed that fields are `required` when they cannot be null and their value can only be
        assigned manually by the user.
        For more information about sobject fields: [1]

        References
        ----------
        [1] :
        https://developer.salesforce.com/docs/atlas.en-us.api.meta/api/sforce_api_objects_list.htm

        """
        if salesforce_object_name not in self._salesforce_object_required_fields_cache:
            self._validate_salesforce_object = False
            response = self.do_request(
                method="GET", path=f"sobjects/{salesforce_object_name}/describe"
            )

            # A required field cannot be null, its value will not be assigned automatically by
            # Salesforce when the record is created, and its value can be assigned by the user.
            self._salesforce_object_required_fields_cache[salesforce_object_name] = [
                fields["name"]
                for fields in json.loads(response)["fields"]
                if not fields["nillable"]
                and not fields["defaultedOnCreate"]
                and fields["createable"]
            ]

        return self._salesforce_object_required_fields_cache[salesforce_object_name]

    @classmethod
    def _obtain_salesforce_object_name_from_path(cls, path: str) -> Optional[str]:
        # Extract Salesforce_object_name taking into account that we'll find something like...
        if "query?q=" in path:
            # ...query?q=SELECT one or more fields FROM an object WHERE filter statements
            query_start_pos = path.find("query?q=") + len("query?q=")
            query = path[query_start_pos:]

            if "WHERE" in query.upper():
                pattern = "FROM\s*(\S*)\s*WHERE"
            else:
                pattern = "FROM\s*(\S*)"

            matches = re.search(pattern, query, re.IGNORECASE)
            if matches:
                salesforce_object_name = matches.group(1)
            else:
                salesforce_object_name = None

        elif "query/" in path:
            # ...query/identifier to get the next rows of a SOQL.
            salesforce_object_name = None

        elif path == "sobjects" or path == "limits":
            salesforce_object_name = None

        elif "sobjects" in path:
            #  ...sobjects/Account/describe, ...sobjects/Account or ...sobjects/Account/ID
            salesforce_object_name = path.split("/")[path.index("sobjects") + 1]
        else:
            salesforce_object_name = None

        return salesforce_object_name

    def _validate_payload_fields(self, payload: Dict[str, str], object_fields: List[str]) -> None:
        for field in payload:
            if field not in object_fields:
                raise SalesforceObjectFieldError(f"`{field}` isn't a valid field.")

    def _validate_required_fields_in_payload(
        self, payload: Dict[str, str], object_required_fields: List[str]
    ) -> None:
        for field in object_required_fields:
            if field not in payload:
                raise SalesforceObjectFieldError(
                    f"`{field}` is a required field and does not appear in the payload."
                )
            else:
                if not payload[field]:
                    raise SalesforceObjectFieldError(
                        f"`{field}` is a required field and must not be empty."
                    )

    @staticmethod
    def _soql_response_to_dataframe(
        response: List[dict],
        column_mapping: Dict[str, str] = None,
        drop_columns_containing: str = None,
    ) -> pd.DataFrame:

        df = pd.json_normalize(data=response)

        if column_mapping:
            df.rename(columns=column_mapping, errors="raise", inplace=True)

        if drop_columns_containing:
            df.drop(
                columns=[column for column in df.columns if drop_columns_containing in column],
                axis=1,
                inplace=True,
            )

        return df

    def do_request(self, method: str, path: str, payload: Optional[Dict[str, str]] = None) -> str:
        """
        Construct and send a request.

        Parameters
        ----------
        method : str {'POST', 'GET', 'PATCH', 'DELETE'}
            An HTTP Request Method.
        path : str
            Relative path of requesting service.
        payload : `Dict[str, str]`, optional
            Payload that contains information that complements the requesting operation.

        Returns
        -------
        str
            Response from Salesforce. This is a JSON encoded as text.

        Raises
        ------
        SalesforceObjectFieldError
            If any field in the payload is invalid, any required field is empty or missing.

        SalesforceObjectError
            If the object in the query is not a valid Salesforce object.

        RequestMethodError
            If the method is not supported, or the payload is missing when needed.

        HTTPError
            If the connection with Salesforce fails, e.g. the requesting resource does not exist.

        """
        if method not in ["POST", "GET", "PATCH", "DELETE"]:
            raise RequestMethodError("Method isn't supported.")

        if not self._is_authenticated:
            self._authenticate()

        if self._validate_salesforce_object:
            path_salesforce_object = self._obtain_salesforce_object_name_from_path(
                path=parse.unquote(path)
            )
            if path_salesforce_object:
                # SOQL is case insensitive, thus comparing in uppercase is fine.
                if path_salesforce_object.upper() not in (
                    object_name.upper() for object_name in self.get_salesforce_object_names()
                ):
                    raise SalesforceObjectError(f"{path_salesforce_object} isn't a valid object")

        url = self._url_to_format.format(
            self._instance_scheme_and_authority, self._api_version, path
        )

        try:
            if method in ["POST", "PATCH"]:
                if not payload:
                    raise RequestMethodError("Payload must be defined for a POST, PATCH request.")

                if self._validate_salesforce_object and path_salesforce_object:
                    object_fields = self.get_salesforce_object_fields(path_salesforce_object)
                    self._validate_payload_fields(payload=payload, object_fields=object_fields)
                    if method == "POST":
                        object_required_fields = self.get_salesforce_object_required_fields(
                            path_salesforce_object
                        )
                        self._validate_required_fields_in_payload(
                            payload=payload,
                            object_required_fields=object_required_fields,
                        )

                # We use reflection to choose which method (POST or PATCH) to execute.
                response = getattr(requests, method.lower())(
                    url=url, json=payload, headers=self.request_headers
                )
            else:  # method == "GET" or method == "DELETE":
                response = getattr(requests, method.lower())(url=url, headers=self.request_headers)

            # Call Response.raise_for_status method to raise exceptions from HTTP errors (e.g. 401
            # Unauthorized).
            response.raise_for_status()
        except requests.exceptions.RequestException:
            raise HTTPError(json.loads(response.text)[0]["message"])
        else:
            self._validate_salesforce_object = True

        # Response is returned as is, it's caller's responsability to do the parsing.
        return response.text

    def do_query_with_SOQL(
        self,
        query: str,
        column_mapping: Dict[str, str] = None,
        drop_columns_containing: str = "attributes",
        drop_empty_columns: bool = False,
    ) -> pd.DataFrame:
        """
        Constructs and sends a request using SOQL.

        Use the Salesforce Object Query Language (SOQL) to search Salesforce data for specific
        information. Once the query is done, the response is flattened and transformed into a pandas
        DataFrame.

        Parameters
        ----------
        query : str
            SOQL with the desired query.

        column_mapping: Dict[str, str], Optional
            Dictionary containing the names of the columns to map, being the key the Salesforce
            Object field name and the value the name of the dataframe. If not given, the Salesforce
            Object field name will be set as column name in the dataframe by default.

        drop_columns_containing: str, Optional
            Columns containing this string will be dropped. By default we drop columns containing
            `attributes`, because SOQL response contains additional `attributes` fields when the
            query contains a selection of a field through relationship.

        drop_empty_columns: bool, Optional
            Columns that do not contain any value will be removed. Set to false by default.

        Returns
        -------
        pd.DataFrame
            A dataframe containing the results of the query.

        Raises
        ------
        requests.exceptions.RequestException
            If the request fails, e.g. the requesting attribute does not exist for the Salesforce
            object class.

        See Also
        --------
        do_request : this method does a request of type GET.

        Notes
        -----
        Use this method when you know which objects the data resides in, and you want to
        retrieve data from a single Salesforce object or from multiple objects that are related.

        The maximum number of records that a single SOQL request can return is 2000, when this limit
        is exceeded, the field `nextRecordsUrl` from the response of the query is used, this method
        already manages this problem, hence all the records of the query will be returned.

        Salesforce object relationships stands for associations between different Salesforce
        objects, when objects are related, their fields can be accessed directly in the SOQL query.

        This method calls the private method _soql_response_to_dataframe. The response JSON object
        is flattened using the Pandas json_normalize method.

        For more information related to SOQL: [1]
        For a complete description of the SOQL syntax: [2]
        For more information about the SOQL requests URIs: [3]
        For more information about the limits of the SOQL requests: [4]
        For more information about Salesforce objects relationships: [5]
        For more information about the json_normalize method: [6]

        References
        ----------
        [1] :
        https://developer.salesforce.com/docs/atlas.en-us.soql_sosl.meta/soql_sosl/sforce_api_calls_soql.htm

        [2] :
        https://developer.salesforce.com/docs/atlas.en-us.soql_sosl.meta/soql_sosl/sforce_api_calls_soql_select.htm

        [3] :
        https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_query.htm

        [4] :
        https://developer.salesforce.com/docs/atlas.en-us.salesforce_app_limits_cheatsheet.meta/salesforce_app_limits_cheatsheet/salesforce_app_limits_platform_soslsoql.htm

        [5] :
        https://developer.salesforce.com/docs/atlas.en-us.api.meta/api/relationships_among_objects.htm

        [6] :
        https://pandas.pydata.org/docs/reference/api/pandas.json_normalize.html

        """
        # The `safe` parameter is set to not escape certain characters. This is done in order
        # to achieve the same behaviour as JavaScript's encodeURIComponent function (which is used
        # by the Salesforce data exporter extension and is the desired behavior to reproduce here)
        # https://stackoverflow.com/a/6618858
        encoded_query = parse.quote(query, safe="()*!'")
        response = self.do_request(method="GET", path=f"query?q={encoded_query}")
        response_dict = json.loads(response)
        records = response_dict["records"]

        # Salesforce returns records in batches. Here we collect all the batches.
        while "nextRecordsUrl" in response_dict:
            next_url = response_dict["nextRecordsUrl"]
            # The nextRecordsUrl field has the form .../query/query_identifier
            query_identifier = next_url.split("/")[-1]
            response = self.do_request(method="GET", path=f"query/{query_identifier}")
            response_dict = json.loads(response)
            records.extend(response_dict["records"])

        df = self._soql_response_to_dataframe(
            response=records,
            column_mapping=column_mapping,
            drop_columns_containing=drop_columns_containing,
        )

        if drop_empty_columns:
            df.dropna(axis=1, how="all", inplace=True)

        return df

    def insert_record(self, sobject: str, payload: Dict[str, str]) -> str:
        """
        Create a new instance of a Salesforce object.

        Create a new record of type `sobject` with the information in the payload.

        Parameters
        ----------
        sobject : str
            A Salesforce object.
        payload : Dict[str, str]
            Payload that contains information to create the record.

        Returns
        -------
        str
            Response from Salesforce. This is a JSON encoded as text.

        See Also
        --------
        do_request : this method does a request of type POST.

        """
        return self.do_request(method="POST", path=f"sobjects/{sobject}", payload=payload)

    def modify_record(self, sobject: str, record_id: str, payload: Dict[str, str]) -> str:
        """
        Update an instance of a Salesforce object.

        Modify a record of type `sobject` with id `record_id` using the new
        information in the `payload`.

        Parameters
        ----------
        sobject : str
            A Salesforce object.
        record_id : str
            The record identifier.
        payload : Dict[str, str]
            Payload that contains information to update the record.

        Returns
        -------
        str
            Response from Salesforce. This is a JSON encoded as text.

        Raises
        ------
        HTTPError
            If the connection with Salesforce fails, e.g. the record does not exist.

        See Also
        --------
        do_request : this method does a request of type PATCH.

        """
        return self.do_request(
            method="PATCH",
            path=f"sobjects/{sobject}/{record_id}",
            payload=payload,
        )

    def delete_record(self, sobject: str, record_id: str) -> str:
        """
        Remove an instance of a Salesforce object.

        Delete a record of type `sobject` with id `record_id`.

        Parameters
        ----------
        sobject : str
            A Salesforce object.
        record_id : str
            The identification of the record.

        Returns
        -------
        str
            Response from Salesforce. This is a JSON encoded as text.

        Raises
        ------
        HTTPError
            If the connection with Salesforce fails, e.g. the record does not exist.

        See Also
        --------
        do_request : this method does a request of type DELETE.

        """
        return self.do_request(method="DELETE", path=f"sobjects/{sobject}/{record_id}")

    def get_remaining_daily_api_requests(self) -> int:
        """
        Return the number of calls still available in the current day.

        Returns
        -------
        int
            The number of remaining daily API requests.

        Raises
        ------
        HTTPError
            If the connection with Salesforce fails, e.g. the record does not exist.

        See Also
        --------
        do_request : this method does a request of type GET.

        Notes
        -----
        For more information about the Salesforce organization limits: [1]

        References
        ----------
        [1] :
        https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_limits.htm

        """
        response = self.do_request(method="GET", path="limits")
        return int(json.loads(response)["DailyApiRequests"]["Remaining"])



from santoku.google.bigquery import BigQueryHandler

"""
This subpackage handles all interactions with Google's cloud services. It makes use of the official
APIs found here: https://github.com/googleapis/google-cloud-python
"""



from typing import Any, Dict, List, Optional, Union

import boto3
from botocore import exceptions


class MessageAttributeError(Exception):
    def __init__(self, message):
        super().__init__(message)


class MessageBatchError(Exception):
    def __init__(self, message):
        super().__init__(message)


class SQSMissingArgumentError(Exception):
    def __init__(self, message):
        super().__init__(message)


class SQSHandler:
    """
    Class to manage operations of Amazon Simple Queue Service (SQS).

    This class is intended to be run on AWS Glue jobs (Python Shell). SQS consists in hosted queues
    that allow sending and receiving messages. This class provides methods to interact with SQS
    queues and makes easy some usual operations. The connection to the SQS queues is done using the
    service class Client of the boto3 library.

    """

    def __init__(self, **kwargs) -> None:
        """
        Instantiate the services classes.

        Parameters
        ----------
        region_name : str
            AWS Region in which to operate the service.

        Notes
        -----
            More information on the available regions: [1]

        References
        ----------
        [1] :
        https://aws.amazon.com/about-aws/global-infrastructure/regions_az/

        """
        self.client = boto3.client(service_name="sqs", **kwargs)
        # Cache to store the url of each queue.
        self.queue_url: Dict[str, str] = {}

    def check_queue_is_fifo(self, queue_name: str) -> bool:
        """
        Check if a queue is of type FIFO.

        Parameters
        ----------
        queue_name : str
            Name of queue.

        Returns
        -------
        bool
            True if the queue is of type FIFO, False if it is standard.

        """
        return queue_name[-5:] == ".fifo"

    def get_queue_url(self, queue_name: str) -> str:
        """
        Return the url of the queue.

        Parameters
        ----------
        queue_name : str
            Name of queue.

        Returns
        -------
        str
            The url of the queue called `queue_name`.

        Raises
        ------
        botocore.exceptions.ClientError
            If the queue does not exist.

        Notes
        -----
        The account url has the form: https://region_name.queue.amazonaws.com/account_number/queue_name
        More information on queue urls: [1]

        References
        ----------
        [1] :
        https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-general-identifiers.html

        """
        response = self.client.get_queue_url(QueueName=queue_name)
        queue_url = response["QueueUrl"]
        return queue_url

    def check_message_attributes_are_well_formed(
        self, message_attributes: Dict[str, Dict[str, str]]
    ) -> None:
        """
        Check whether the message attributes are correct.

        Check if the message attributes are following the structure that AWS SQS requires. A boolean
        will be returned together with an error message in case it is not correctly structured.

        Parameters
        ----------
        message_attributes : Dict[str, Dict[str, str]]
            Message attributes to send.

        Returns
        -------
        Tuple[bool, str]
            True and an empty message will be returned if the message attributes are correclty
            formed. False and an error message explaining the error will be returned otherwise.

        Raises
        ------
        MessageAttributeError
            If the message attributes are not correclty structured.

        Notes
        -----
        A message attribute can be of `DataType` Binary, Number and String.
        The `TypeValue` of the attribute must be StringValue if `DataType` is String or Number, and
        BinaryValue if `DataType` is Binary. StringListValues and BinaryListValues are
        options that should be accepted in the future but are not implemented yet. Strings are
        Unicode with UTF-8 binary encoding. Each message can have up to 10 message attributes.
        More information on the sqs message attributes: [1]

        References
        ----------
        [1] :
        https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-message-attributes.html

        """

        if len(message_attributes) > 10:
            error_message = "Messages can have up to 10 attributes."
            raise MessageAttributeError(error_message)

        for _, attribute_content in message_attributes.items():

            if not isinstance(attribute_content, dict):
                error_message = "Each message attribute must be a dictionary containing 'DataType' and 'StringValue' arguments."
                raise MessageAttributeError(error_message)

            if "DataType" not in attribute_content:
                error_message = "'DataType' argument is missing in message attribute."
                raise MessageAttributeError(error_message)

            if attribute_content["DataType"] not in ["Binary", "Number", "String"]:
                error_message = (
                    "The supported types for 'DataType' argument are: Binary, Number and String."
                )
                raise MessageAttributeError(error_message)

            if attribute_content["DataType"] == "String" and "StringValue" not in attribute_content:
                error_message = (
                    "'StringValue' argument is required for message attributes of type String."
                )
                raise MessageAttributeError(error_message)

            if attribute_content["DataType"] == "Number" and "StringValue" not in attribute_content:
                error_message = (
                    "'StringValue' argument is required for message attributes of type Number."
                )
                raise MessageAttributeError(error_message)

            if attribute_content["DataType"] == "Binary" and "BinaryValue" not in attribute_content:
                error_message = (
                    "'BinaryValue' argument is required for message attributes of type Binary."
                )
                raise MessageAttributeError(error_message)

        return None

    def send_message(
        self,
        queue_name: str,
        message_body: str,
        message_deduplication_id: str = None,
        message_group_id: str = None,
        message_attributes: Dict[str, dict] = {},
    ) -> Dict[str, Any]:
        """
        Deliver a message to a SQS Queue.

        A message with `message_body` and optionally `message_attributes` will be sent to the queue
        named `queue_name`.

        Parameters
        ----------
        queue_name : str
            Name of the queue to send the message.
        message_body : str
            Body of of the message to be sent.
        message_deduplication_id : str, optional
            Token used for deduplication of sent messages. Required for FIFO queues.
        message_group_id : str, optional
             Tag that specifies that a message belongs to a specific message group. Required for
             FIFO queues.
        message_attributes : Dict[str, Dict[str, Any]], optional
            Attributes of the message to send. Will be empty by default.

        Returns
        -------
        Dict[str, Any]
            A response containing information about messages sent. In these information we can find
            the message id that identifies the message, and metadata containing the HTTP status code
            of each message.

        Raises
        ------
        MessageAttributeError
            If the message argument is not correctly structured.

        SQSMissingArgumentError
            If`queue_name` is a FIFO queue and the message doesn't contain a `MessageDeduplicationId`
            or a MessageGroupId

        See Also
        --------
        get_queue_url : this method retrieves the queue url with the given queue name.
        check_message_attributes_are_well_formed : this method checks the given message attritutes.

        Notes
        -----
        The maximum size that a message can have is 256KB (message body and attributes all together).

        """

        # Update the urls if not registered yet.
        if queue_name in self.queue_url:
            queue_url = self.queue_url[queue_name]
        else:
            queue_url = self.get_queue_url(queue_name=queue_name)
            self.queue_url[queue_name] = queue_url

        # Check if the type of queue is FIFO
        fifo = self.check_queue_is_fifo(queue_name)

        # Add optional arguments
        args: Dict[str, Union[str, Dict[str, dict]]] = {}

        if message_attributes:
            # Check whether the message attributes are correctly structured.
            self.check_message_attributes_are_well_formed(message_attributes=message_attributes)
            args["MessageAttributes"] = message_attributes
        if fifo:
            if not message_deduplication_id:
                raise SQSMissingArgumentError(
                    "Missing `message_deduplication_id` argument for FIFO Queues"
                )

            if not message_group_id:
                raise SQSMissingArgumentError("Missing `message_group_id` argument for FIFO Queues")

            args["MessageDeduplicationId"] = message_deduplication_id
            args["MessageGroupId"] = message_group_id

        response = self.client.send_message(
            QueueUrl=queue_url,
            MessageBody=message_body,
            **args,
        )

        return response

    def send_message_batch(self, queue_name: str, entries: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Deliver a list of messages to a SQS Queue.

        A list of batch request `entries` containing details of the messages will be sent to the
        queue named `queue_name`.

        Parameters
        ----------
        queue_name : str
            Name of the queue to send the batch of messages.
        entries : List[Dict[str, Any]]
            List of messages containing a unique message Id, a message body and message attributes
            optionally.

        Returns
        -------
        Dict[str, Any]
            A response containing information about messages sent. In this information
            we can find the message id that identifies each message, and metadata containing the
            HTTP status code of each message.

        Raises
        ------
        MessageAttributeError
            If any message does not contain id or message body, if the ids are not unique, or if the
            message attributes are not correctly structured.

        SQSMissingArgumentError
            If `queue_name` is a FIFO queue and any message doesn't contain a `MessageDeduplicationId`
            or a MessageGroupId

        See Also
        --------
        get_queue_url : this method retreives the queue url with the given queue name.
        check_message_attributes_are_well_formed : this method checks the message attributes given.

        Notes
        -----
        A batch can contain up to 10 messages.
        The maximum size that the batch of messages can have is 256KB, this means the sum of all
        messages (message body and attributes all together) cannot exceed 256KB.

        """

        if not entries:
            error_message = "The list of 'entries' cannot be emtpy."
            raise MessageBatchError(error_message)

        if len(entries) > 10:
            error_message = "The maximum number of messages allowed in a batch is 10."
            raise MessageBatchError(error_message)

        # Check if the type of queue is FIFO
        fifo = self.check_queue_is_fifo(queue_name)

        # Check if all messages have the required attributes and if the message attributes are
        # correctly structured.
        entries_ids = []
        for entry in entries:

            if "Id" not in entry:
                error_message = "'Id' attribute is required for each message."
                raise MessageBatchError(error_message)
            entries_ids.append(entry["Id"])

            if "MessageBody" not in entry:
                error_message = "'MessageBody' attribute is required for each message."
                raise MessageBatchError(error_message)

            if "MessageAttributes" in entry:
                self.check_message_attributes_are_well_formed(entry["MessageAttributes"])

            if fifo:
                if "MessageDeduplicationId" not in entry:
                    raise SQSMissingArgumentError(
                        "`MessageDeduplicationId` is required for each message for FIFO Queues"
                    )

                if "MessageGroupId" not in entry:
                    raise SQSMissingArgumentError(
                        "`MessageGroupId` is required for each message for FIFO Queues"
                    )

        # Check the message ids are unique along the batch.
        if len(entries_ids) != len(set(entries_ids)):
            error_message = "'ID' attribute must be unique along all the messages."
            raise MessageBatchError(error_message)

        # Update the urls if not registered yet.
        if queue_name in self.queue_url:
            queue_url = self.queue_url[queue_name]
        else:
            queue_url = self.get_queue_url(queue_name=queue_name)
            self.queue_url[queue_name] = queue_url

        response = self.client.send_message_batch(QueueUrl=queue_url, Entries=entries)
        return response

    def receive_message(self, queue_name: str) -> Dict[str, Any]:
        """
        Retrieves multiple messages from the SQS Queue.

        Parameters
        ----------
        queue_name : str
            Name of the queue to receive a message.

        Returns
        -------
        Dict[str, Any]
            A response containing information about the received messages. In these information we
            can find the message id that identifies each message, the receipt handle used to delte
            the message, and the content of the message such as the message body and message
            attributes.

        Raises
        ------
        MessageAttributeError
            If any message does not contain id or message body, if the ids are not unique, if the
            message attributes are not correctly structured.

        See Also
        --------
        get_queue_url : this method retreives the queue url with the given queue name.
        check_message_attributes_are_well_formed : this method checks the message attributes given.

        Notes
        -----
        The maximum number of messages we can receive in each call is 10.
        There is an attribute called visibility timeout, that is the time that SQS keep a message
        invisible after it has been received once. After that period of time the message will be
        visible again. This is why removing the message after receiving it is so important.

        References
        ----------
        [1] :
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.receive_message

        """
        # Update the urls if not registered yet.
        if queue_name in self.queue_url:
            queue_url = self.queue_url[queue_name]
        else:
            queue_url = self.get_queue_url(queue_name=queue_name)
            self.queue_url[queue_name] = queue_url

        response = self.client.receive_message(QueueUrl=queue_url, MaxNumberOfMessages=10)
        return response

    def delete_message(self, queue_name: str, receipt_handle: str) -> None:
        """
        Remove a message from the queue.

        A message with the `receipt_handle` will be deleted from the queue with name `queue_name`.

        Parameters
        ----------
        queue_name : str
            Name of the queue to delete the message from.
        receipt_handle : str
            Identifier obtained from receiving the message to delete.

        Returns
        -------
        None

        See Also
        --------
        get_queue_url : this method retreives the queue url with the given queue name.

        Notes
        -----
        You must always receive a message before you can delete it (you can't put a message into the
        queue and then recall it.
        The difference between the receipt handle and the message id is that the first one is
        associated with the action of receiving the message and not the message itself.
        More information on the receipt handle: [1]

        References
        ----------
        [1]:
        https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-general-identifiers.html

        """
        # Update the urls if not registered yet.
        if queue_name in self.queue_url:
            queue_url = self.queue_url[queue_name]
        else:
            queue_url = self.get_queue_url(queue_name=queue_name)
            self.queue_url[queue_name] = queue_url

        self.client.delete_message(QueueUrl=queue_url, ReceiptHandle=receipt_handle)
        return None



import os

import boto3
import pytest
from botocore import exceptions
from moto import mock_sqs
from santoku.aws.sqs import (
    MessageAttributeError,
    MessageBatchError,
    SQSHandler,
    SQSMissingArgumentError,
)


@pytest.fixture(scope="class")
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "eu-west-1"


@pytest.fixture(scope="class")
def sqs_handler(aws_credentials):
    with mock_sqs():
        sqs_handler = SQSHandler()
        yield sqs_handler


@pytest.fixture(scope="function")
def standard_queue(sqs_handler, request):
    queue_name = "test_standard_queue"
    sqs_client = sqs_handler.client
    try:
        sqs_client.get_queue_url(QueueName=queue_name)
    except exceptions.ClientError:
        # By catching the exception we check the queue certainly does not exist.
        pass
    else:
        raise EnvironmentError(f"{queue_name} should not exist.")
    sqs_client.create_queue(QueueName=queue_name)
    yield queue_name

    def teardown():
        queue_url = sqs_client.get_queue_url(QueueName=queue_name)["QueueUrl"]
        sqs_client.delete_queue(QueueUrl=queue_url)

    request.addfinalizer(teardown)


@pytest.fixture(scope="function")
def fifo_queue(sqs_handler, request):
    queue_name = "test_fifo_queue.fifo"
    sqs_client = sqs_handler.client
    try:
        sqs_client.get_queue_url(QueueName=queue_name)
    except exceptions.ClientError:
        pass
    else:
        raise EnvironmentError(f"{queue_name} should not exist.")
    sqs_client.create_queue(QueueName=queue_name, Attributes={"FifoQueue": "true"})
    yield queue_name

    def teardown():
        queue_url = sqs_client.get_queue_url(QueueName=queue_name)["QueueUrl"]
        sqs_client.delete_queue(QueueUrl=queue_url)

    request.addfinalizer(teardown)


@pytest.fixture(scope="function")
def message_attributes():
    return {
        "StringAttribute": {
            "DataType": "String",
            "StringValue": "Test string value",
        },
        "NumberAttribute": {
            "DataType": "Number",
            "StringValue": "1000000000000000",
        },
        "BinaryAttribute": {
            "DataType": "Binary",
            "BinaryValue": "Test binary value",
        },
    }


class TestSQSHandler:
    def test_check_queue_existence(self, sqs_handler, standard_queue):
        # Test an existing queue. Success expected.
        assert sqs_handler.get_queue_url(queue_name=standard_queue)

        # Test a non-existent queue (according to boto3 documentation, this function only throws
        # ClientError exception when the queue is not found. ClientError is a generic exception, but
        # we dont match this specific message because boto changes this frequently and we would
        # easily break compatibility). Failure expected.
        with pytest.raises(exceptions.ClientError):
            assert sqs_handler.get_queue_url(queue_name="WRONG_QUEUE_NAME")

    def test_check_queue_is_fifo(self, sqs_handler, standard_queue, fifo_queue):
        # Test a fifo queue. Success expected.
        assert sqs_handler.check_queue_is_fifo(queue_name=fifo_queue)

        # Test a standard queue. Failure expected.
        assert not sqs_handler.check_queue_is_fifo(queue_name=standard_queue)

    def test_get_queue_url(self, sqs_handler, standard_queue):
        # Test getting the name of a queue that does exist. Success expected.
        obtained_url = sqs_handler.get_queue_url(queue_name=standard_queue)
        region = os.environ["AWS_DEFAULT_REGION"]

        # Moto uses the following account id by default. Currently programatic access is not allowed,
        # there is an issue in Moto's github repository: https://github.com/spulec/moto/issues/850.
        # Currently Moto is supposed to allow mocking entire accounts but it’s bugged (check the
        # follow link: https://github.com/spulec/moto/issues/2634). We should be patching when
        # mocking accounts/profiles becomes stable.
        moto_aws_account = "123456789012"

        expected_url = f"https://{region}.queue.amazonaws.com/{moto_aws_account}/{standard_queue}"
        assert obtained_url == expected_url

        # Test getting the name of a queue that does not exist (according to boto3 documentation,
        # this function only throws ClientError exception when the queue is not found. ClientError
        # is a generic exception, but we dont match this specific message because boto changes this
        # frequently and we would easily break compatibility). Failure expected.
        with pytest.raises(exceptions.ClientError):
            sqs_handler.get_queue_url(queue_name="WRONG_QUEUE_NAME")

    def test_check_message_attributes_are_well_formed(self, sqs_handler, message_attributes):
        # Message attributes correctly structured. Success expected.
        sqs_handler.check_message_attributes_are_well_formed(message_attributes)

        # Message with more than 10 attributes. Failure expected.
        num_attributes = len(message_attributes)
        max_num_attributes = 10
        message_attribute_content = message_attributes["StringAttribute"]

        # Create 11 attributes.
        attributes = message_attributes.copy()
        for i in range(num_attributes + 1, max_num_attributes + 2):
            attributes[f"TestAttribute{i}"] = message_attribute_content

        expected_message = "Messages can have up to 10 attributes."
        with pytest.raises(MessageAttributeError, match=expected_message):
            sqs_handler.check_message_attributes_are_well_formed(attributes)

        # Message attributes not correctly structured. Failure expected.
        attributes = {"WrongAttribute": "WrongValue"}
        expected_message = "Each message attribute must be a dictionary containing 'DataType' and 'StringValue' arguments."
        with pytest.raises(MessageAttributeError, match=expected_message):
            sqs_handler.check_message_attributes_are_well_formed(attributes)

        # Message attributes that does not contain the required arguments. Failure
        # expected.
        attributes = {"WrongAttribute": {"StringValue": "Test string Value"}}
        expected_message = "'DataType' argument is missing in message attribute."
        with pytest.raises(MessageAttributeError, match=expected_message):
            sqs_handler.check_message_attributes_are_well_formed(attributes)

        attributes = {"WrongAttribute": {"DataType": "String", "BinaryValue": "Test string value"}}
        expected_message = (
            "'StringValue' argument is required for message attributes of type String."
        )
        with pytest.raises(MessageAttributeError, match=expected_message):
            sqs_handler.check_message_attributes_are_well_formed(attributes)

        attributes = {"WrongAttribute": {"DataType": "Number", "BinaryValue": "1"}}
        expected_message = (
            "'StringValue' argument is required for message attributes of type Number."
        )
        with pytest.raises(MessageAttributeError, match=expected_message):
            sqs_handler.check_message_attributes_are_well_formed(attributes)

        attributes = {"WrongAttribute": {"DataType": "Binary", "StringValue": "Test binary value"}}
        expected_message = (
            "'BinaryValue' argument is required for message attributes of type Binary."
        )
        with pytest.raises(MessageAttributeError, match=expected_message):
            sqs_handler.check_message_attributes_are_well_formed(attributes)

    def test_send_message_to_standard_queue(self, sqs_handler, standard_queue, message_attributes):
        # Send a message to a standard queue without message attributes. Success expected.
        message_body = "Test message body."
        response = sqs_handler.send_message(
            queue_name=standard_queue,
            message_body=message_body,
            message_attributes=message_attributes,
        )
        assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

        # Check if the received ids match the sent ones.
        expected_message_id = response["MessageId"]

        queue_url = sqs_handler.client.get_queue_url(QueueName=standard_queue)["QueueUrl"]
        response = sqs_handler.client.receive_message(QueueUrl=queue_url)
        obtained_message_id = response["Messages"][0]["MessageId"]
        assert obtained_message_id == expected_message_id

        # Check if the received message contains the message attributes
        assert "MessageAttributes" in response["Messages"][0]

    def test_send_message_to_fifo_queue(self, sqs_handler, fifo_queue, message_attributes):
        # Send a message to a FIFO queue. Success expected.
        message_body = "Test message body."
        message_deduplication_id = "test_deduplication_id"
        message_group_id = "test_group_id"

        response = sqs_handler.send_message(
            queue_name=fifo_queue,
            message_body=message_body,
            message_deduplication_id=message_deduplication_id,
            message_group_id=message_group_id,
            message_attributes=message_attributes,
        )
        assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

        # Check if the received ids match the sent ones.
        expected_message_id = response["MessageId"]

        queue_url = sqs_handler.client.get_queue_url(QueueName=fifo_queue)["QueueUrl"]
        response = sqs_handler.client.receive_message(QueueUrl=queue_url)
        obtained_message_id = response["Messages"][0]["MessageId"]
        assert obtained_message_id == expected_message_id

        # Check if the received message contains the message attributes
        assert "MessageAttributes" in response["Messages"][0]

        # Send a message to a FIFO queue without passing the required attirbutes. Failure expected.
        with pytest.raises(SQSMissingArgumentError):
            sqs_handler.send_message(
                queue_name=fifo_queue,
                message_body=message_body,
            )

    def test_send_message_batch_to_standard_queue(
        self, sqs_handler, standard_queue, message_attributes
    ):
        entries = []

        # Send a batch without messages to a standard queue. Failure expected.
        error_message = "The list of 'entries' cannot be emtpy."
        with pytest.raises(MessageBatchError, match=error_message):
            sqs_handler.send_message_batch(
                queue_name=standard_queue,
                entries=entries,
            )

        # Send a batch of n messages to a standard queue. Success expected.
        number_messages = 3
        for i in range(1, number_messages + 1):
            message_id = f"ID{i}"
            message_body = f"Test message body {i}."
            message = {
                "Id": message_id,
                "MessageBody": message_body,
                "MessageAttributes": message_attributes,
            }
            entries.append(message)

        response = sqs_handler.send_message_batch(
            queue_name=standard_queue,
            entries=entries,
        )
        assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

        # Check whether the received message ids coincide with the ones sent.

        # Collect the message ids.
        expected_message_ids = set()
        for message_response in response["Successful"]:
            expected_message_ids.add(message_response["MessageId"])

        # We use sets because the reception order is not guaranteed.
        obtained_message_ids = set()
        queue_url = sqs_handler.client.get_queue_url(QueueName=standard_queue)["QueueUrl"]

        # Set the maximum number of messages, by default it was 1.
        response = sqs_handler.client.receive_message(
            QueueUrl=queue_url, MaxNumberOfMessages=number_messages
        )

        for message in response["Messages"]:
            obtained_message_ids.add(message["MessageId"])
        assert obtained_message_ids == expected_message_ids

        # Send a batch with more than the maximum allowed. Failure expected.
        max_num_messages = 10
        for i in range(number_messages + 1, max_num_messages + 2):
            message_id = f"ID{i}"
            message_body = f"Test message body {i}."
            message = {
                "Id": message_id,
                "MessageBody": message_body,
                "MessageAttributes": message_attributes,
            }
            entries.append(message)

        error_message = "The maximum number of messages allowed in a batch is 10."
        with pytest.raises(MessageBatchError, match=error_message):
            sqs_handler.send_message_batch(
                queue_name=standard_queue,
                entries=entries,
            )

        # Send a batch with repeated ids. Failure expected.
        entries = []
        message_id = "ID1"
        message_body = "Test message body."
        message = {
            "Id": message_id,
            "MessageBody": message_body,
            "MessageAttributes": message_attributes,
        }
        entries.append(message)
        entries.append(message)

        error_message = "'ID' attribute must be unique along all the messages."
        with pytest.raises(MessageBatchError, match=error_message):
            sqs_handler.send_message_batch(
                queue_name=standard_queue,
                entries=entries,
            )

    def test_send_message_batch_to_fifo_queue(self, sqs_handler, fifo_queue, message_attributes):
        entries = []

        # Send a batch without messages to a standard queue. Failure expected.
        error_message = "The list of 'entries' cannot be emtpy."
        with pytest.raises(MessageBatchError, match=error_message):
            sqs_handler.send_message_batch(
                queue_name=fifo_queue,
                entries=entries,
            )

        # Send a batch of n messages to a standard queue. Success expected.
        number_messages = 3
        for i in range(1, number_messages + 1):
            message_id = f"ID{i}"
            message_body = f"Test message body {i}."
            message = {
                "Id": message_id,
                "MessageBody": message_body,
                "MessageAttributes": message_attributes,
                "MessageDeduplicationId": f"test_message_deduplication_id_{i}",
                "MessageGroupId": "test_message_group_id",
            }
            entries.append(message)

        response = sqs_handler.send_message_batch(
            queue_name=fifo_queue,
            entries=entries,
        )
        assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

        # Check whether the received message ids coincide with the ones sent.

        # Collect the message ids.
        expected_message_ids = set()
        for message_response in response["Successful"]:
            expected_message_ids.add(message_response["MessageId"])

        # We use sets because the reception order is not guaranteed.
        obtained_message_ids = set()
        queue_url = sqs_handler.client.get_queue_url(QueueName=fifo_queue)["QueueUrl"]

        # Set the maximum number of messages, by default it was 1.
        response = sqs_handler.client.receive_message(
            QueueUrl=queue_url, MaxNumberOfMessages=number_messages
        )

        for message in response["Messages"]:
            obtained_message_ids.add(message["MessageId"])
        assert obtained_message_ids == expected_message_ids

        # Send a batch of n messages to a FIFO queue without all the required attributes. Failure
        # expected.
        number_messages = 3
        for i in range(1, number_messages + 1):
            message_id = f"ID{i}"
            message_body = f"Test message body {i}."
            message = {
                "Id": message_id,
                "MessageBody": message_body,
                "MessageAttributes": message_attributes,
            }
            entries.append(message)

        with pytest.raises(SQSMissingArgumentError):
            sqs_handler.send_message_batch(
                queue_name=fifo_queue,
                entries=entries,
            )

    def test_receive_message(self, sqs_handler, standard_queue, message_attributes):
        # Send a message to a standard queue. Success expected.
        message_body = "Test message body."
        sqs_handler.send_message(
            queue_name=standard_queue,
            message_body=message_body,
            message_attributes=message_attributes,
        )

        response = sqs_handler.receive_message(queue_name=standard_queue)
        message = response["Messages"][0]
        obtained_message_body = message["Body"]
        obtained_message_attributes = message["MessageAttributes"]
        assert obtained_message_body == message_body
        assert len(obtained_message_attributes) == len(message_attributes)
        for (
            obtained_attribute_name,
            obtained_attribute_content,
        ) in obtained_message_attributes.items():
            # In the received message attributes the binary values are transformed into binary
            if obtained_attribute_content["DataType"] == "Binary":
                assert (
                    obtained_attribute_content["BinaryValue"].decode("utf8")
                    == message_attributes[obtained_attribute_name]["BinaryValue"]
                )
            else:
                assert obtained_attribute_content == message_attributes[obtained_attribute_name]

    def test_delete_message(self, sqs_handler, standard_queue, message_attributes):
        # Read a message after it is deleted in the queue. Failure expected.
        message_body = "Test message body."
        sqs_handler.send_message(
            queue_name=standard_queue,
            message_body=message_body,
            message_attributes=message_attributes,
        )

        # Get the receipt handler.
        queue_url = sqs_handler.client.get_queue_url(QueueName=standard_queue)["QueueUrl"]
        response = sqs_handler.client.receive_message(QueueUrl=queue_url)
        message = response["Messages"][0]
        receipt_handle = message["ReceiptHandle"]

        sqs_handler.delete_message(queue_name=standard_queue, receipt_handle=receipt_handle)

        # Read the messages in the queue.
        queue_url = sqs_handler.client.get_queue_url(QueueName=standard_queue)["QueueUrl"]
        response = sqs_handler.client.receive_message(QueueUrl=queue_url)
        assert "Messages" not in response



import boto3

from abc import ABC, abstractmethod
from typing import Any, Dict, Generator

from botocore import client


"""
Class to manage generic methods in AWS.

This class contains methods with generic syntax, the same method works for more than on service.
"""


def paginate(
    client: client, method: str, **kwargs: Dict[str, Any]
) -> Generator[Dict[str, Any], None, None]:
    """
    Iterates over the pages of an API operation results.

    Paginators act as an abstraction over the process of iterating over an entire result set of
    a truncated API operation. Yields an iterable with the response obtained from applying
    `method`.

    Parameters
    ----------
    method : str
        Name of the API operation request.
    kwargs : Dict[str, Any]
        Additional arguments for the specified method.

    Yields
    ------
    Generator[Dict[str, Any], None, None]
        Responses dictionaries of the `method`.

    Notes
    -----
    More information on the use of AWS paginators: [1].

    References
    ----------
    [1] :
    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/paginators.html

    """
    paginator = client.get_paginator(operation_name=method)
    for page in paginator.paginate(**kwargs):
        yield page



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['santoku',
 'santoku.aws',
 'santoku.aws.test',
 'santoku.google',
 'santoku.google.test',
 'santoku.salesforce',
 'santoku.salesforce.test',
 'santoku.slack',
 'santoku.slack.test',
 'santoku.sql',
 'santoku.sql.test',
 'santoku.utils',
 'santoku.utils.test']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.16.50,<2.0.0',
 'jsonschema>=3.2.0,<4.0.0',
 'mysql-connector-python>=8.0.5,<9.0.0',
 'pandas-gbq>=0.17.1,<0.18.0',
 'pandas>=1.2.4,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'slackclient>=2.5.0,<3.0.0',
 'tldextract>=3.1.0,<4.0.0']

setup_kwargs = {
    'name': 'santoku',
    'version': '220408.33',
    'description': 'Custom Python wrapper around many third party APIs, including AWS, BigQuery, Slack and Salesforce.',
    'long_description': '[![PyPI version](https://badge.fury.io/py/santoku.svg)](https://badge.fury.io/py/santoku)\n\n[![Install deps, Test & Release](https://github.com/wiris/santoku/actions/workflows/cd.yml/badge.svg)](https://github.com/wiris/santoku/actions/workflows/cd.yml)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n# What is Santoku?\n\nSantoku is a toolkit written in Python for interacting with AWS, Google Cloud platform, Salesforce and Slack.\n\nThe purpose of Santoku is to have the interactions with all the external services collected in a single package. The package contains wrappers around the respective APIs and high level methods for the most common patterns in order to simplify the interaction with those services, whether by being shorter to type, more descriptive, more specific to our needs or simply easier to read for developers.\n\n## Quickstart\n\n### Installation\n\nIf you have the wheel, you can install it with:\n\n```bash\npip install --upgrade --force-reinstall dist/santoku-*.whl\n```\n\nRun the following command to install it from PyPI:\n\n```bash\npip install santoku\n```\n\nOr use the following to install it via Poetry\n\n```bash\npoetry add santoku\n```\n\n### How To Use It\n\nYou can use the package as follows:\n\n```python\nfrom santoku.slack.slack_bot_handler import SlackBotHandler\n\nslack_bot = SlackBotHandler.from_aws_secrets_manager("your_secret")\n```\n\n## Content\n\nThe package `santoku` contains several subpackages: `aws`, `google`, `salesforce`, `slack`, `sql`. Each subpackage provides connection to different external services and are formed by a collection of modules, where each module consists of handlers for more specific services. Each handler class has unit tests to ensure the correct behaviour of the methods of these classes.\n\n### AWS\n\nAWS (Amazon Web Services) is a cloud computing platform that provides a set of primitive abstract technical infrastructure and distributed computing building blocks and tools.\n\nThe connection to AWS has been done through the AWS SDK, which in Python is called [boto3](https://github.com/boto/boto3). We provide wrappers of the `boto3` SDK to make easy the operations to interact with different services.\n\nThe use of this subpackage requires having AWS credentials somewhere. We provide flexibility to either keep credentials in AWS credentials/configuration file, set environment variables, or to pass them directly as arguments in the initializer of each handler class. More info on AWS configurations and credentials [here](https://docs.aws.amazon.com/cli/latest/userguide/cli-chap-configure.html).\n\nThe unit tests in this subpackage implement mocks to the AWS services and do not pretended to access or modify the environment of your real account. In order to have safer unit tests for the AWS subpackage, we use moto, a mocking library for most AWS services, which allows our methods to interact with a fully mocked version of the AWS environment via decorators while not needing an actual connection to the internet or a test AWS account.\n\n#### Amazon S3\n\nAmazon Simple Storage Service (Amazon S3) is an object storage service that offers scalability, data availability, security, and performance.\n\nWe provide methods to easily list and delete objects inside buckets; read and write content within S3 objects; upload a dataframe into csv or parket format to a specific location; generate and upload an Amazon Quicksight manifest in S3 in order to create analysis in Amazon Quicksight, and so on.\n\nAn object can be uploaded to S3 with the following:\n\n```python\nfrom santoku.aws.s3_handler import S3Handler\n\ns3_handler = S3Handler()\ns3_handler.put_object(bucket="your_bucket_name", object_key="your_object_key", content="Your object content.")\n```\n\n#### AWS Secrets Manager\n\nAWS Secrets Manager protects secrets needed to access applications, services, and IT resources. The service allows rotating, managing, and retrieving credentials, keys, and other secrets.\n\nWe provide methods to get the content of a previously created secret.\n\n```python\nfrom santoku.aws.secrets_manager_handler import SecretsManagerHandler\n\nsecrets_manager = SecretsManagerHandler()\nsecret_content = secrets_manager.get_secret_value(secret_name="your_secret_name")\n```\n\nWe use this service as our default credential manager. Most classes that require some form of authentication in santoku are provided with alternative class methods that retrieve the credentials directly from Secrets Manager. For example, instead of directly providing credentials to the BigQuery handling class, we simply provide it with the name of the secret where they are stored:\n\n```python\nfrom santoku.google.bigquery import BigQueryHandler\n\nbigquery_handler = BigQueryHandler(\n    type="your_type",\n    project_id="your_project_id"\n    private_key_id="your_private_key_id"\n    private_key="your_private_key"\n    client_email="your_client_email"\n    client_id="your_client_id"\n    auth_uri="your_auth_uri"\n    token_uri="your_token_uri"\n    auth_provider_x509_cert_url="your_auth_provider_x509_cert_url"\n    client_x509_cert_url="your_client_x509_cert_url"\n)\n```\n\nor\n\n```python\nbigquery_handler = BigQueryHandler.from_aws_secrets_manager(\n    secret_name="your_secret_name"\n)\n```\n\n#### Amazon Simple Queue Service\n\nAmazon Simple Queue Service (SQS) is a fully managed message queuing service that supports programmatic sending of messages via web service applications as a way to communicate over the Internet.\n\nWe provide methods to receive, delete, and send single or a batch of messages.\n\n```python\nfrom santoku.aws.sqs_handler import SQSHandler\n\nsqs_handler = SQSHandler()\nentries = [\n    {\n        "Id": "Id1",\n        "MessageBody": "Your message 1",\n    },\n    {\n        "Id": "Id2",\n        "MessageBody": "Your message 1",\n    }\n]\nsqs_handler.send_message_batch(queue_name="your_queue_name", entries=entries)\n```\n\n### Google Cloud Platform\n\nGoogle Cloud Platform a suite of cloud computing services provided by Google that runs on the same Cloud infrastructure that Google uses internally for its end-user products.\n\nThe connection to Google Cloud Platform has been done using the [google-cloud-core](https://googleapis.dev/python/google-api-core/latest/index.html) package.\n\nThe use of this subpackage requires having [Google Cloud Platform credentials](https://cloud.google.com/docs/authentication/production#obtaining_and_providing_service_account_credentials_manually) (in this case, a service account for programmatic access), these can be passed as arguments in the initializer of the handler class directly, or you can store them in AWS Secrets Manager and retrieve them during the initialization using the class method instead.\n\nWe provide a handler that allows doing queries on BigQuery services:\n\n```python\nquery_results = bigquery_handler.get_query_results(query="SELECT * FROM `your_table`")\n```\n\n### Salesforce\n\nSalesforce is a Customer Relationship Management (CRM) platform that gives to the marketing, sales, commerce, and service depertments a single, shared view of every customer.\n\nThe connection to Salesforce has been done using the [Salesforce REST API](https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/quickstart.htm).\n\nThe use of this subpackage requires having Salesforce credentials, these can be passed as arguments in the initializer of the handler class directly, or you can store them in AWS Secrets Manager and retrieve them during the initialization using the class method instead.\n\nThis subpackage provide methods to insert/modify/delete salesforce object records. You can perform operations by doing HTTP requests directly or using methods with higher level of abstraction, which are easier to handle. The lasts ones are just wrappers of the HTTP request method. To obtain records you can perform queries using SOQL.\n\nThe unit tests require valid Salesforce credentials to be executed. The tests are implemented in the way that no new data will remain in the account and no existent data will be modified. However, having Salesforce credentials for sandbox use is recommended.\n\nYou can use the package to perform a request as follows.\n\n```python\nfrom santoku.salesforce.objects_handler import ObjectsHandler\n\nobjects_handler = ObjectsHandler(\n    auth_url="your_auth_url",\n    username="your_username",\n    password="your_password",\n    client_id="your_client_id",\n    client_secret="your_client_secret",\n)\ncontact_payload = {"FirstName": "Alice", "LastName": "Ackerman", "Email": "alice@example.com"}\n\nobjects_handler.do_request(method="POST", path="sobjects/Contact", payload=contact_payload)\n```\n\nor insert a record with\n\n```python\nobjects_handler.insert_record(sobject="Contact", payload=contact_payload)\n```\n\nFinally, you can do a SOQL query with:\n\n```python\nrecords = objects_handler.do_query_with_SOQL("SELECT Id, Name from Contact")\n```\n\n### Slack\n\nSlack is a proprietary business communication platform. A Slack Bot is a nifty way to run code and automate tasks. In Slack, a bot is controlled programmatically via a bot user token that can access one or more of Slack’s APIs.\n\nThe connection to Slack has been done using the [Slack Web API](https://slack.dev/python-slackclient/basic_usage.html).\n\nThe use of this subpackage requires having Slack API Token of a Slack Bot, which can be passed as argument in the initializer of the handler class directly, or you can store it in AWS Secrets Manager and retrieve it during the initialization using the class method instead.\n\nThis subpackage provide methods to send messages to a channel. A message can be sent with:\n\n```python\nfrom santoku.slack.slack_bot_handler import SlackBotHandler\n\nslack_bot_handler = SlackBotHandler(api_token="your_api_token")\nslack_bot_handler.send_message(channel="your_chanel_name", message="Your message.")\n```\n\n### SQL\n\nSQL (Structured Query Language) is a domain-specific language designed for managing data held in a relational database management system (RDBMS). The purpose of this subpackage is to provide connection to different RDBMSs.\n\n#### MySQL\n\nMySQL is an open-source RDBMS. The connection to MySQL has been done using the [MySQL Connector for python](https://dev.mysql.com/doc/connector-python/en/).\n\nThe use of this subpackage requires having MySQL authentication parameters, which can be passed as argument in the initializer of the handler class directly, or you can store it in AWS Secrets Manager and retrieve it during the initialization using the class method instead.\n\nThis subpackage provides methods to do queries and retrieve the results in different forms.\n\n```python\nfrom santoku.sql.mysql_handler import MySQLHandler\n\nmysql_handler = MySQLHandler(user="your_user", password="your_password", host="your_host", database="your_database")\nmysql_handler.get_query_results(query="SELECT * FROM your_table")\n```\n\n## Development\n\n### Project\n\nWe use Poetry to handle this project. Poetry is a tool for dependency management and packaging in Python. It allows you to declare the libraries your project depends on and it will manage (install/update) them for you. More detais in [their documentation](https://python-poetry.org/docs/basic-usage/).\n\nPoetry is already included in the development environment.\n\n#### Dependencies\n\nIf you want to add dependencies to your project, you can specify them in the `tool.poetry.dependencies` section of the `pyproject.toml` file. Also, instead of modifying the `pyproject.toml` file by hand, you can use the add command:\n\n```bash\npoetry add <package_name>\n```\n\nPoetry will automatically find a suitable version constraint and install the package and subdependencies.\n\n**Note:** Remember to commit changes of `poetry.lock` and `pyproject.toml` files after adding a new dependency.\n\n**Note:** You can find more details on how to handle versions of packages [here](https://python-poetry.org/docs/dependency-specification/).\n\n### Environment\n\nWe provide a development environment that uses the Visual Studio Code Remote - Containers extension. This extension lets you use a Docker container in order to have a consistent and easily reproducible development environment.\n\nThe files needed to build the container are located in the `.devcontainer` directory:\n\n* `devcontainer.json` contains a set of configurations, tells VSCode how to access the container and which extensions it should install.\n* `Dockerfile` defines instructions for the building of the container image.\n\nMore info [here](https://code.visualstudio.com/docs/remote/containers-tutorial).\n\n#### Environment Variables\n\nThe containerized environment will automatically set as environment variables the your variables stored in a `.env` file at the top level of the repository. For example, this is required for certain tests that require credentials, which are (of course) not versioned. Be aware that:\n\n* The Docker image building process will **fail** if you do not include a `.env` file at the top level of the repository.\n* If you change the contents of the `.env` file you will need to rebuild the container for the changes to take effect within the environment.\n\n#### Sharing Git credentials with your container\n\nThe containerized environment will automatically forward your local SSH agent if one is running. More info [here](https://code.visualstudio.com/docs/remote/containers#_using-ssh-keys). It works for Windows and Linux.\n\n### Creating a PR\n\nCreate a pull request (PR) to propose and collaborate on changes to the project. These changes MUST BE proposed in a branch, which ensures that the main branch only contains finished and approved work.\n\nBe sure to run tests locally before commiting your changes.\n\n#### Running tests\n\nThe tests are implemented with pytest and there are unit tests for each of the handler modules. Tests in the `aws` subpackage implement mocks to S3 and do not require real credentials, however, the remaining tests in other subpackages do.\n\nTo run the tests just execute `pytest` if you are already inside Poetry virtual environment or `poetry run pytest`.\n\nMoreover, when a PR is created a GitHub Actions CI pipeline (see [`.github/workflows/ci.yml`](./.github/workflows/ci.yml)) is executed. This pipeline is in charge of running tests.\n\n### Release\n\nWheel is automatically created and uploaded to PyPI by the GitHub Actions CD pipeline (see [`.github/workflows/cd.yml`](./.github/workflows/cd.yml)) when the PR is merged in main branch.\n\n\n## Why Santoku?\n\nFrom Wikipedia:\n\n```text\nThe Santoku bōchō (Japanese: 三徳包丁; "three virtues" or "three uses") or Bunka bōchō (文化包丁) is a general-purpose kitchen knife originating in Japan. Its blade is typically between 13 and 20 cm (5 and 8 in) long, and has a flat edge and a sheepsfoot blade that curves down an angle approaching 60 degrees at the point. The term Santoku may refer to the wide variety of ingredients that the knife can handle: meat, fish and vegetables, or to the tasks it can perform: slicing, chopping and dicing, either interpretation indicating a multi-use, general-purpose kitchen knife.\n```\n',
    'author': 'Data Science Team at WIRIS',
    'author_email': 'info@wiris.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/wiris/santoku',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



import os
import json

import pytest
import boto3

from base64 import b64encode

from moto import mock_secretsmanager

from santoku.aws.secretsmanager import SecretsManagerError, SecretsManagerHandler


@pytest.fixture(scope="class")
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "eu-west-1"


@pytest.fixture(scope="class")
def secrets_manager_handler(aws_credentials):
    with mock_secretsmanager():
        secrets_manager = SecretsManagerHandler()
        yield secrets_manager


@pytest.fixture(scope="class")
def secret_content():
    username = "test_user"
    password = "test_password"
    secret_content = {"username": username, "password": password}
    return secret_content


@pytest.fixture(scope="function")
def string_secret(secrets_manager_handler, secret_content, request):
    secret_name = "test/secret_name"
    secrets_manager_handler.client.create_secret(
        Name=secret_name, SecretString=json.dumps(secret_content)
    )
    yield secret_name

    def teardown():
        secrets_manager_handler.client.delete_secret(
            SecretId=secret_name, ForceDeleteWithoutRecovery=True
        )

    request.addfinalizer(teardown)


@pytest.fixture(scope="function")
def binary_secret(secrets_manager_handler, secret_content, request):
    secret_name = "test/secret_name"
    secret_binary = b64encode(json.dumps(secret_content).encode())
    secrets_manager_handler.client.create_secret(Name=secret_name, SecretBinary=secret_binary)
    yield secret_name

    def teardown():
        secrets_manager_handler.client.delete_secret(
            SecretId=secret_name, ForceDeleteWithoutRecovery=True
        )

    request.addfinalizer(teardown)


class TestSecretsManagerHandler:
    def test_get_string_secret(self, secrets_manager_handler, secret_content, string_secret):
        # Retrieve a string secret. Success expected.
        obtained_secret = secrets_manager_handler.get_secret_value(secret_name=string_secret)
        assert obtained_secret["username"] == secret_content["username"]
        assert obtained_secret["password"] == secret_content["password"]

    def test_get_binary_secret(self, secrets_manager_handler, secret_content, binary_secret):
        # Retrieve a string secret. Success expected.
        obtained_secret = secrets_manager_handler.get_secret_value(secret_name=binary_secret)
        assert obtained_secret["username"] == secret_content["username"]
        assert obtained_secret["password"] == secret_content["password"]

    def test_get_non_existent_secret(self, secrets_manager_handler):
        # Retreive a secret that does not exist. Failure expected.
        expected_message = "Secrets Manager can't find the resource you asked for."
        with pytest.raises(SecretsManagerError, match=expected_message) as e:
            obtained_secret = secrets_manager_handler.get_secret_value(secret_name="wrong_secret")

    # TODO: Test with an KMS encripted key.
